
/*
 * FSDB_GO/src/fsdb.go
 * Copyright (c) 2016-2017 John Heidemann
 * All rights reserved.                                            
 *                                                                
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 * 
 */


// Package fsdb implements input and output for Fsdb files
// as defined at https://www.isi.edu/~johnh/SOFTWARE/FSDB/.
package fsdb

import (
	"bufio"
	"errors"
	"fmt"
        "io"
	"reflect"
	"strconv"
	"strings"
	"unicode"
)

// An Fsdb encapsulates an Fsdb things common to I and O.
type Fsdb struct {
	headerrow string
	headerSet bool
	fscode string
	fscodeFieldsFunc func(rune) bool
	fs string
	rscode string
	rs string
	compression string
	Empty string
	cols []string
	colsToI map[string]int
}

// A Reader encapsulates an Fsdb-specific stream input.
type FsdbReader struct {
	Fsdb
	inR io.ReadCloser
	in *bufio.Reader
	commentFunc func(string)
}

// A Writer encapsulates an Fsdb-specific stream output.
type FsdbWriter struct {
	Fsdb
	outW io.WriteCloser
	out *bufio.Writer
}

/***********************************************************************/

// Fsdb.init(): interal setup
func (f *Fsdb)init() {
	f.fscode = "s"
	f.fs = " "
	f.rscode = "D"
	f.Empty = "-"
	f.colsToI = make(map[string]int)
}

// NewFsdb: Create an Fsdb object (could be reader or writer)
func NewFsdb() *Fsdb {
	f := new(Fsdb)
	f.init()
	return f
}

// internalColCreate: bookkeeping for new columns, either from a header or ColCreate
func (f *Fsdb) internalColCreate(col string) error {
	if f.headerSet {
		return errors.New("attempt to add column to frozen fsdb: " + col)
	}
	if _, ok := f.colsToI[col]; ok {
		return errors.New("duplicate col definition: " + col)
	}
	// establishNewColMapping
	i := len(f.cols)
	f.cols = append(f.cols, col)
	f.colsToI[col] = i
	f.colsToI[strconv.Itoa(i)] = i
	return nil
}

// Fsdb.ColToI: see if a col is in the table, returning its INDEX and OK if present
func (f *Fsdb) ColToI(col string) (int, bool) {
	i, ok := f.colsToI[col]
	return i, ok
}

// parseFSCode: take the -F x into whatever internal magic we need.
// FSCodes:
//    D (default): any number of spaces
//    s: single space
//    t: single tab
//    D (default): any number of spaces
func (f *Fsdb) parseFSCode(value string) error {
	f.fscode = value
	switch f.fscode {
	case "D": // default
		f.fscodeFieldsFunc = unicode.IsSpace
		f.fs = "  "
	case "s": // single space
		f.fscodeFieldsFunc = func(r rune) bool { return r == ' '; }
		f.fs = " "
	case "t": // single tab
		f.fscodeFieldsFunc = func(r rune) bool { return r == '\t'; }
		f.fs = "\t"
	default:
		if f.fscode[0] == 'X' || f.fscode[0] == 'C' {
			return errors.New("FS codes Xxx and Cc not yet supported in FSDB_GO")
		}
		return errors.New("unknown FS code " + f.fscode)
	}
	return nil
}

// parseRSCode: take the -R x into whatever internal magic we need.
func (f *Fsdb) parseRSCode(value string) error {
	return errors.New("RS codes not yet supported in FSDB_GO")
}

/***********************************************************************/

// NewFsdbReader builds a Fsdb reader around an io.Reader.
//
// xxx: needs to take options.
func NewFsdbReader(inR io.ReadCloser) (rp *FsdbReader, err error) {
	br := bufio.NewReader(inR)
	r := new(FsdbReader)
	r.init()
	r.inR = inR
	r.in = br
	if r.headerrow == "" {
		if err = r.readHeaderrow(); err != nil {
			return
		}
		if err = r.parseHeaderrow(); err != nil {
			return
		}
	}
	if  r.headerrow == "" {
		return nil, errors.New("no header line")
	}
	r.headerSet = true  // go read-only
	rp = r
	return
}

// parseRSCode: take the -R x into whatever internal magic we need.
func (r *FsdbReader) Close() error {
	r.in = nil
	err := r.inR.Close()
	if err != nil {
		r.inR = nil
	}
	return err
}

// readHeaderrow: read and save a header.
func (r *FsdbReader) readHeaderrow() error {
	h, err := r.in.ReadString('\n')
	if err != nil {
		return err
	}
	r.headerrow = h
	return nil
}

// parseHeaderrow: setup the Reader from a header (either read or provided).
func (r *FsdbReader) parseHeaderrow() error {
	f := strings.Fields(r.headerrow)
	if len(f) < 1 {
		return errors.New("invalid fsdb header: no fields")
	}
	if f[0] == "#h"  || f[0] == "#L" {
		return errors.New("fsdb-go does not support fsdb 1.x headers")
	}
	if f[0] != "#fsdb" {
		// xxx: could support old #h
		return errors.New("invalid fsdb header: no #fsdb magic number")
	}
	f = f[1:]
	
	// options
	for f[0][0] == '-' {
		key := f[0]
		if len(f) < 2 {
			return errors.New("invalid fsdb header: flag " + f[0] + " with no argument")
		}
		value := f[1]
		f = f[2:]
		switch key[1] {
		case 'F':
			if err := r.parseFSCode(value); err != nil {
				return err
			}
		case 'R':
			if err := r.parseRSCode(value); err != nil {
				return err
			}
		case 'Z':
			return errors.New("internal compression not yet supported in FSDB_GO")
		default:
			return errors.New("unknown key in header line " + key)
		}
	}
	
	// now only fields are left
	for _, col := range f {
		if err := r.internalColCreate(col); err != nil {
			return err
		}
	}

	return nil
}

// r.ReadRow: read a new Fsdb row and returns the fields as a slice of strings.
func (r *FsdbReader) ReadRow() ([]string, error) {
	if !r.headerSet {
		return nil, errors.New("header not yet set")
	}
	for {
		ln, err := r.in.ReadString('\n')
		if err != nil {
			// propagate EOF or other errors
			return nil, err
		}
		if ln[0] == '#' {
			if r.commentFunc != nil {
				r.commentFunc(ln)
			}
			continue
		}
		if ln[len(ln)-1] == '\n' {
			ln = ln[0:len(ln)-1] // chomp!
		}
		f := strings.FieldsFunc(ln, r.fscodeFieldsFunc)
		return f, nil
	}
	// notreached
}

const skippedFsdbStructElem = -1

// FsdbStructElem: state about demarshalling one (structure / record) element.
type FsdbStructElem struct {
	// i int // implicit
	inF int  // -1 if field is skipped
	base int // only for integers
};

// FsdbStructElem: state about demarshalling from a reader to a structure
type FsdbReaderState struct {
	r *FsdbReader
	empty string
	lastUsedColI int
	elems []FsdbStructElem
}

// MakeReadStructState: save state to support Reader-to-struct demarshalling.
// The structure is given in outPtr, with demarshalling based on tagKey
// in that struct's tags.
func (r *FsdbReader) MakeReadStructState(outPtr interface{}, tagKey string) (s *FsdbReaderState, err error) {
	var frs FsdbReaderState
	frs.r = r
	frs.empty = r.Empty
	// out must be a pointer to a struct, indirect to the struct
	out := reflect.ValueOf(outPtr).Elem()
	frs.elems = make([]FsdbStructElem, out.NumField())
//	fmt.Printf("found %d fields\n", len(frs.elems))
	for i := 0; i < len(frs.elems); i++ {
		valueField := out.Field(i)
		typeField := out.Type().Field(i)
		tag := typeField.Tag
//		fmt.Printf("%d: name: %s, tag: %s\n", i, typeField.Name, tag)
//		fmt.Printf("%d: name: %s, value: %v, tag: %s\n",
//			i, typeField.Name, valueField.Interface(), tag)
		colName := tag.Get(tagKey)
		if colName == "" {
			// don't fill in this one
			frs.elems[i].inF = skippedFsdbStructElem
			continue
		}
		colI, ok := r.ColToI(colName)
		if !ok {
			return nil, fmt.Errorf("FsdbReader.MakeReadStructState: field %s requires column %s, that is missing from input", typeField.Name, colName)
		}
		frs.elems[i].inF = colI
		if colI > frs.lastUsedColI {
			frs.lastUsedColI = colI
		}
		switch valueField.Type().Kind() {
		case reflect.Int, reflect.Int8, reflect.Int16,
			reflect.Int32, reflect.Int64, 
			reflect.Uint, reflect.Uint8, reflect.Uint16, 
			reflect.Uint32, reflect.Uint64:
			base, hasBase := tag.Lookup("base")
			if hasBase {
				frs.elems[i].base, err = strconv.Atoi(base)
				if err != nil {
					return nil, err
				}
			} else {
				frs.elems[i].base = 10
			}
		}
	}
	return &frs, nil
}

// ReadWithStructState: read one row into a structure using saved state
func (r *FsdbReader) ReadWithStructState(outPtr interface{}, frs *FsdbReaderState) error {
	row, err := frs.r.ReadRow();
	if err != nil {
		return err;
	}
	// out must be a pointer to a struct, indirect to the struct
	out := reflect.ValueOf(outPtr).Elem()
	n_elems := out.NumField()
	if n_elems != len(frs.elems) {
		return fmt.Errorf("FsdbReader.readWithStructState: suprising number of elems, expected %d got %d", len(frs.elems), n_elems)
	}
	// makes sure enough input
	for len(row) < frs.lastUsedColI {
		row = append(row, frs.empty)
	}
	for i := 0; i < len(frs.elems); i++ {
		valueField := out.Field(i)
		// typeField := out.Type().Field(i)
		var rawResult string
		if frs.elems[i].inF != skippedFsdbStructElem {
			rawResult = row[frs.elems[i].inF]
		} else {
			rawResult = ""
		}
		switch valueField.Type().Kind() {
		case reflect.Int, reflect.Int8, reflect.Int16,
			reflect.Int32, reflect.Int64:
			var result int64
			if rawResult != "" {
				result, err = strconv.ParseInt(rawResult, frs.elems[i].base, 64)
				if err != nil {
					return err
				}
			} else {
				result = 0
			}
			valueField.SetInt(result)
		case reflect.Uint, reflect.Uint8, reflect.Uint16,
			reflect.Uint32, reflect.Uint64:
			var result uint64
			if rawResult != "" {
				result, err = strconv.ParseUint(rawResult, frs.elems[i].base, 64)
				if err != nil {
					return err
				}
			} else {
				result = 0
			}
			valueField.SetUint(result)
		case reflect.Float32, reflect.Float64:
			var result float64
			if rawResult != "" {
				result, err = strconv.ParseFloat(rawResult, 64)
				if err != nil {
					return err
				}
			} else {
				result = 0.0
			}
			valueField.SetFloat(result)
		case reflect.String:
			valueField.SetString(rawResult)
		default:
			return fmt.Errorf("unsupported kind: %s", valueField.Type().Kind())
		}
	}
	return nil
}

// ReadStructFunc: a thunk we return to do the reading work.
type ReadStructFunc func(interface {}) error;

// MakeReadStructFunc: generate a ReadStructFunc to demarshall a given Reader and structure.
func (r *FsdbReader) MakeReadStructFunc(outPtr interface{}, tagKey string) (rsf ReadStructFunc, err error) {
	var frs *FsdbReaderState
	if frs, err = r.MakeReadStructState(outPtr, tagKey); err != nil {
		return nil, err
	}
	rsf = func (rsfOutPtr interface{}) error {
		return r.ReadWithStructState(rsfOutPtr, frs)
	}
	return rsf, nil
}

/***********************************************************************/

// NewFsdbWriter: create a writer around some kind of io.WriteCloser
func NewFsdbWriter(ioW io.WriteCloser) (*FsdbWriter, error) {
	bw := bufio.NewWriter(ioW)
	w := new(FsdbWriter)
	w.init()
	w.outW = ioW
	w.out = bw

	return w, nil
}


// FsdbWriter.Close(): done with it
func (w *FsdbWriter) Close() error {
	if err := w.out.Flush(); err != nil {
		return err
	}
	w.out = nil
	if err := w.outW.Close(); err != nil {
		return err
	}
	w.outW = nil
	return nil
}

// FsdbWriter.SetFSCode: set the field separator from a code
func (w *FsdbWriter) SetFSCode(fs string) error {
	return w.parseFSCode(fs)
}

// FsdbWriter.SetFSCode: set the record separator
func (w *FsdbWriter) SetRSCode(rs string) error {
	return w.parseRSCode(rs)
}

// FsdbWriter.SetCompression: not yet implemented
func (w *FsdbWriter) SetCompression(rs string) error {
	return errors.New("internal compression not yet supported in FSDB_GO")
}

// FsdbWriter.SetEmpty: set the empty value
func (w *FsdbWriter) SetEmpty(e string) error {
	w.Empty = e
	// xxx: could check to see if it overlaps with the field sepearator
	return nil
}

// FsdbWriter.AddCol: add a new COLUMN to the output
func (w *FsdbWriter) AddCol(col string) error {
	return w.internalColCreate(col)
}

// FsdbWriter.AddCols: add a slice of COLUMNS to the output
func (w *FsdbWriter) AddCols(cols []string) error {
	for _, col := range cols {
		if err := w.internalColCreate(col); err != nil {
			return err
		}
	}
	return nil
}

// FsdbWriter.updateHeaderrow: make the headerrow match the internal state
func (w *FsdbWriter) updateHeaderrow() {
	var s string
	s = "#fsdb ";
	if w.fscode != "" && w.fscode != "D" {
		s += "-F " + w.fscode + " ";
	}
	if w.rscode != "" && w.rscode != "D" {
		s += "-R " + w.fscode + " ";
	}
	if w.compression != "" && w.compression != "none" {
		s += "-Z " + w.compression + " ";
	}
	s +=  strings.Join(w.cols, " ")
	w.headerrow = s
}

// FsdbWriter.Finalize: freeze things and output the header row
func (w *FsdbWriter) Finalize() error {
	if w.headerSet {
		return errors.New("attempt to finalize an already finalized writer")
	}
	if w.rscode != "" && w.rscode != "D" {
		return errors.New("attempt to finalize with a record separator; not yet supported")
	} else {
		w.rs = "\n"  // set the default
	}
	w.updateHeaderrow()
	if _, err := w.out.WriteString(w.headerrow + "\n"); err != nil {
		return err
	}
	w.headerSet = true
	return nil
}

// w.WriteRow writes a new Fsdb row, given as a slice of strings
func (w *FsdbWriter) WriteRow(row []string) error {
	if !w.headerSet {
		return errors.New("write attempt to non-finalized FsdbWriter")
	}
	// xxx: convert string to bytes?
	_, err := w.out.WriteString(strings.Join(row, w.fs) + w.rs)
	return err
}

// w.WriteComment writes a comment (no "# " or newline)
func (w *FsdbWriter) WriteComment(s string) error {
	if !w.headerSet {
		return errors.New("WriteComment attempt to non-finalized FsdbWriter")
	}
	// xxx: convert string to bytes?
	_, err := w.out.WriteString("# " + s + "\n")
	return err
}

// w.WriteString writes a string (just a pass-through)
func (w *FsdbWriter) WriteString(s string) error {
	if !w.headerSet {
		return errors.New("WriteComment attempt to non-finalized FsdbWriter")
	}
	_, err := w.out.WriteString(s)
	return err
}

